import com.cycling74.max.*;
import com.cycling74.msp.*;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.RandomAccessFile;

import javazoom.jl.decoder.Bitstream;
import javazoom.jl.decoder.Decoder;
import javazoom.jl.decoder.Header;
import javazoom.jl.decoder.JavaLayerException;
import javazoom.jl.decoder.SampleBuffer;

public class mp3play extends MSPPerformer {
Thread t;
String filename;
Header currFrame;

private Bitstream bitstream;
private Decoder decoder;
private RandomAccessFileInputStream in = null;

short[] sampsL = new short[44100];
short[] sampsR = new short[44100];

private float filePosMS = 0;
private int recPos = 0;
private int playPos = 0;
private int playing = 0;
private boolean looping = false;

private static final String[] INLET_ASSIST = new String[] { "messages in" };

private static final String[] OUTLET_ASSIST = new String[] { "(signal) Channel 1 Output", "(signal) Channel 2 Output", "File position (ms)", "Info Outlet" };

public mp3play() {
declareInlets(new int[] { DataTypes.ALL });
declareOutlets(new int[] { SIGNAL, SIGNAL, DataTypes.FLOAT, DataTypes.ALL });

setInletAssist(INLET_ASSIST);
setOutletAssist(OUTLET_ASSIST);

createInfoOutlet(false);

t = new Thread() {
public void run() {
while (1 == 1) {
while (playing == 1) {
int diff = 0;
if (recPos < playPos)
diff = recPos - playPos + sampsL.length;
else
diff = recPos - playPos;

if (diff + decoder.getOutputBlockSize() <= sampsL.length / 2) {
try {
short[] frame = decodeFrame();

if (frame != null) {
{
for (int i = 0; i < frame.length / 2; i++) {
sampsL[recPos] = frame[i * 2];
sampsR[recPos] = frame[i * 2 + 1];

recPos++;
recPos %= sampsL.length;
}
filePosMS += currFrame.ms_per_frame();

}
} else {
if (looping) {
seek(0);
} else
playing = 0;
}

} catch (JavaLayerException e) {
}
}

outlet(2, filePosMS);

try {
Thread.sleep(20);
} catch (InterruptedException e) {
}
}

try {
Thread.sleep(20);
} catch (InterruptedException e) {
}

}
}
};

t.start();
}

public void open(String name) {
in = null;
playing = 0;
try {
in = new RandomAccessFileInputStream(new File(name));

} catch (FileNotFoundException e) {
outlet(3, new Atom[] { Atom.newAtom("error"), Atom.newAtom("filenotfound") });
}

if (in != null) {
filePosMS = 0;
recPos = 0;
playPos = 0;
filename = name;
File f = new File(filename);
decoder = new Decoder();
bitstream = new Bitstream(in);
} else
playing = 0;
}

public void inlet(int i) {
int inlet = getInlet();
if (inlet == 0) {
if (in != null) {
playing = i;
}
}
}

public void seek(float endPosMS) throws JavaLayerException {
playing = 0;
recPos = 0;
playPos = 0;
if (currFrame != null && in != null) {
int offset = (int) (endPosMS / currFrame.ms_per_frame()) * currFrame.calculate_framesize();
try {
in.seek((long) (offset));
filePosMS = currFrame.total_ms(offset);
} catch (IOException e) {
}
}

decoder = new Decoder();
bitstream = new Bitstream(in);
playing = 1;
}

public void perform(MSPSignal[] in, MSPSignal[] out) {
float[] L = out[0].vec;
float[] R = out[1].vec;

if (playing == 1) {
int diff = 0;
if (recPos < playPos)
diff = recPos - playPos + sampsL.length;
else
diff = recPos - playPos;

if (diff > 0) {
for (int i = 0; i < L.length; i++) {
L[i] = (float) sampsL[playPos] / (float) Short.MAX_VALUE;
R[i] = (float) sampsR[playPos] / (float) Short.MAX_VALUE;

playPos += 1;
if (playPos >= sampsL.length)
playPos = 0;
}
} else {
for (int i = 0; i < L.length; i++) {
L[i] = R[i] = 0;
}
}
} else {
for (int i = 0; i < L.length; i++) {
L[i] = R[i] = 0;
}
}
}

protected short[] decodeFrame() throws JavaLayerException {
try {
if (decoder == null)
return null;
if (bitstream == null)
return null;

Header h = bitstream.readFrame();
if (h == null) {
outlet(3, "done");
return null;
} else
currFrame = h;

SampleBuffer output = (SampleBuffer) decoder.decodeFrame(currFrame, bitstream);
short[] samps = output.getBuffer();
bitstream.closeFrame();

return samps;
} catch (RuntimeException e) {
throw new JavaLayerException("Exception decoding audio frame", e);
}
}

public void info() {
if (in != null) {
File f = new File(filename);
float lengthMS = currFrame.total_ms((int) f.length());
int frequency = currFrame.frequency();
int bitRate = currFrame.bitrate() / 1000;
outlet(3, new Atom[] { Atom.newAtom("filename"), Atom.newAtom(filename) });
outlet(3, new Atom[] { Atom.newAtom("samplerate"), Atom.newAtom(frequency) });
outlet(3, new Atom[] { Atom.newAtom("bitrate"), Atom.newAtom(bitRate + "Kbps") });
outlet(3, new Atom[] { Atom.newAtom("length"), Atom.newAtom(lengthMS) });
} else
outlet(3, new Atom[] { Atom.newAtom("error"), Atom.newAtom("nofileopen") });
}

public void loop(int i) {
if (i == 1)
looping = true;
else
looping = false;
}

public void notifyDeleted() {
t.stop();
t = null;
}
}


